import json
from collections import OrderedDict
from typing import Any, Optional

import dash
from dash import Input, Output, State, clientside_callback, html, page_container
from dash.development.base_component import Component

"""
This module holds utilities related to the [Dash pages](https://dash.plotly.com/urls).
"""

_ID_CONTENT = "_component_content"
_COMPONENT_PATH_REGISTRY: dict[Component, list[str]] = OrderedDict()
_PROP_PATH_REGISTRY: dict[Component, dict[str, list[str]]] = OrderedDict()
_CONTAINER_REGISTRY: dict[Component, Component] = {}
_COMPONENT_CONTAINER = html.Div(id=_ID_CONTENT, disable_n_clicks=True, style=dict(display="contents"))

# region Monkey patch page registration function

_original_register_page = dash.register_page


def _register_page(*args, page_components=None, page_properties=None, **kwargs):
    _original_register_page(*args, **kwargs)
    # Resolve page.
    module = kwargs["module"] if "module" in kwargs else args[0]
    page = dash.page_registry[module]
    # Register callbacks for page props.
    if page_properties is not None:
        for component in page_properties:
            _set_props(component, page["path"], page_properties[component])
    # Resolve any page components.
    if page_components is None:
        return
    for component in page_components:
        _set_visible(component, page["path"])


dash.register_page = _register_page


# endregion

# region Public interface


def set_page_container_style_display_contents():
    """
    Changes the style of the page container (and the page content container) so that their children are rendered
    as if they were children of the page container's parent (see https://caniuse.com/css-display-contents). This is
    an advantage if you are using css grid, as it makes it possible to mix the page components with other components.
    """
    page_container.style = dict(display="contents")
    for child in page_container.children:
        if child.id == "_pages_content":
            child.style = dict(display="contents")


def set_default_container(container: Component):
    """
    Per default, page components are rendered into the '_COMPONENT_CONTAINER' declared above.
    Use this function to change the default container.
    :param container: the container into which page components will be rendered by default
    :return: None
    """
    global _COMPONENT_CONTAINER
    _COMPONENT_CONTAINER = container


def assign_container(component: Component, container: Component):
    """
    By default, page components are rendered into the '_COMPONENT_CONTAINER' declared above. Call this function to
    specify that the component should be rendered in a different container.
    :param component: the (page) component in question
    :param container: the container into which the component will be rendered
    :return: None
    """
    if component in _CONTAINER_REGISTRY:
        raise ValueError("You can assign a component to one container.")
    _CONTAINER_REGISTRY[component] = container


def _set_visible(component: Component, path: str):
    """
    Register path(s) for which a component should be visible.
    :param component: the (page) component in question
    :param path: the (url) path for which the component should be visible
    :return: None
    """
    _COMPONENT_PATH_REGISTRY.setdefault(component, []).append(path)


def _set_props(component: Component, path: str, prop_map: dict[str, Any]):
    """
    Register path(s) for which a particular props should be set.
    :param component: the (page) component in question
    :param path: the (url) path for which the props should be set
    :param prop_map: the props, i.e. (prop name, prop value) pairs
    :return: None
    """
    for prop in prop_map:
        _PROP_PATH_REGISTRY.setdefault(component, OrderedDict()).setdefault(prop, {})[path] = prop_map[prop]


def setup_page_components() -> html.Div:
    """
    Initializes the page components and returns the (default) container into which the components are rendered.
    :return: the default container, into which page components are rendered. Must be included in the layout,
    unless all (page) components are assigned to custom containers (via 'assign_container')
    """
    _setup_callbacks()
    return _COMPONENT_CONTAINER


# endregion

# region Utils


def _prepare_container(container: Optional[Component] = None):
    container = _COMPONENT_CONTAINER if container is None else container
    # Make sure children is a list.
    if container.children is None:
        container.children = []
    if not isinstance(container.children, list):
        container.children = [container.children]
    return container


def _setup_callbacks():
    store = dash.dash._ID_STORE
    location = dash.dash._ID_LOCATION
    # Setup callbacks for page components.
    components = list(_COMPONENT_PATH_REGISTRY.keys())
    for component in components:
        # Wrap in div container, so we can hide it.
        cid = component._set_random_id()
        wrapper = html.Div(
            component,
            disable_n_clicks=True,
            style=dict(display="none"),
            id=f"{cid}_wrapper",
        )
        # Add to container.
        container = _prepare_container(_CONTAINER_REGISTRY.get(component, _COMPONENT_CONTAINER))
        container.children.append(wrapper)
        # Setup callback.
        f = f"""function(y, x){{
            const paths = {_COMPONENT_PATH_REGISTRY[component]};
            if(paths.includes(x)){{
                return {{display: "contents"}};
            }}
            return {{display: "none"}};
        }}"""
        clientside_callback(
            f,
            Output(wrapper, "style", allow_duplicate=True),
            Input(store, "data"),
            State(location, "pathname"),
            prevent_initial_call="initial_duplicate",
        )
    # Setup callbacks for page props.
    components = list(_PROP_PATH_REGISTRY.keys())
    for component in components:
        for prop in _PROP_PATH_REGISTRY[component]:
            path_map = _PROP_PATH_REGISTRY[component][prop]
            default = getattr(component, prop, None)
            # Setup callback.
            f = f"""function(y, x){{
                const path_map = JSON.parse(\'{json.dumps(path_map)}\');
                if (x in path_map){{
                    return path_map[x];
                }}
                return JSON.parse(\'{json.dumps(default)}\');
            }}"""
            clientside_callback(
                f,
                Output(component, prop, allow_duplicate=True),
                Input(store, "data"),
                State(location, "pathname"),
                prevent_initial_call="initial_duplicate",
            )


# endregion



from typing import Union

from pydantic import BaseModel


def sse_message(data: str = "[DONE]") -> str:
    """
    Given a string, formats it as an SSE event.
    """
    return f"data: {data}\n\n"


def sse_options(payload: Union[str, BaseModel], **kwargs):
    """
    Given a string or Pydantic base model, returns a dictionary that can be used as options for StreamingBuffer.
    """
    options = {"payload": payload}
    # For Pydantic models, convert to JSON and set the content type.
    if isinstance(payload, BaseModel):
        options = {
            "payload": payload.model_dump_json(),
            "headers": {"Content-Type": "application/json"},
        }
    # Merge with additional options.
    return {**options, **kwargs}



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class EventSource(Component):
    """An EventSource component.
An interface to server sent events in Dash

Keyword arguments:

- id (string; optional):
    The ID used to identify this component in Dash callbacks.

- close (boolean; optional):
    Close event source.

- error (string; optional):
    Error.

- message (string; optional):
    Received message.

- readyState (number; optional):
    A number representing the state of the connection. Possible values
    are CONNECTING (0), OPEN (1), or CLOSED (2).

- url (string; required):
    A DOMString representing the URL of the source.

- withCredentials (boolean; optional):
    A boolean value indicating whether the EventSource object was
    instantiated with cross-origin (CORS) credentials set (True), or
    not (False, the default)."""
    _children_props = []
    _base_nodes = ['children']
    _namespace = 'dash_extensions'
    _type = 'EventSource'
    @_explicitize_args
    def __init__(self, id=Component.UNDEFINED, close=Component.UNDEFINED, error=Component.UNDEFINED, message=Component.UNDEFINED, readyState=Component.UNDEFINED, withCredentials=Component.UNDEFINED, url=Component.REQUIRED, **kwargs):
        self._prop_names = ['id', 'close', 'error', 'message', 'readyState', 'url', 'withCredentials']
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'close', 'error', 'message', 'readyState', 'url', 'withCredentials']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs and excess named props
        args = {k: _locals[k] for k in _explicit_args}

        for k in ['url']:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')

        super(EventSource, self).__init__(**args)



from __future__ import annotations

import dataclasses
import functools
import hashlib
import inspect
import json
import logging
import secrets
import struct
import sys
import threading
import uuid
from collections import defaultdict
from datetime import datetime
from itertools import compress
from typing import Any, Callable, Dict, Generic, List, Optional, Tuple, TypeVar, Union

import dash
import plotly

# Enable enrich as drop-in replacement for dash
# noinspection PyUnresolvedReferences
from dash import (  # lgtm [py/unused-import]; noqa: F401
    ALL,
    ALLSMALLER,
    MATCH,
    ClientsideFunction,
    Input,
    Output,
    State,
    callback_context,  # noqa: F401
    ctx,  # noqa: F401
    dash_table,  # noqa: F401
    dcc,
    development,  # noqa: F401
    exceptions,  # noqa: F401
    html,
    no_update,
    page_container,  # noqa: F401
    page_registry,  # noqa: F401
    register_page,  # noqa: F401
    resources,  # noqa: F401
)
from dash._callback_context import context_value
from dash._utils import patch_collections_abc
from dash.dependencies import DashDependency, _Wildcard  # lgtm [py/unused-import]
from dash.development.base_component import Component
from dataclass_wizard import asdict, fromdict
from flask import session
from flask_caching.backends import FileSystemCache, RedisCache
from pydantic import BaseModel  # type: ignore

from dash_extensions import CycleBreaker

T = TypeVar("T")

_wildcard_mappings = {ALL: "<ALL>", MATCH: "<MATCH>", ALLSMALLER: "<ALLSMALLER>"}
_wildcard_values = list(_wildcard_mappings.values())

DEPENDENCY_APPEND_PREFIX = "dash_extensions_"


# region DependencyCollection


def build_index(structure, entry, index):
    if isinstance(structure, list):
        for i, s in enumerate(structure):
            build_index(s, entry + [i], index)
        return index
    if isinstance(structure, dict):
        for k in structure:
            build_index(structure[k], entry + [k], index)
        return index
    if isinstance(structure, DashDependency):
        index.append(entry)
        return index
    raise ValueError(f"Unsupported structure {str(structure)}")


def validate_structure(structure, level=0):  # noqa: C901
    if isinstance(structure, DashDependency):
        if level == 0:
            return [structure]
        return structure
    if isinstance(structure, tuple):
        result = list(structure)
        for i, entry in enumerate(result):
            result[i] = validate_structure(entry, level=level + 1)
        return result
    if isinstance(structure, list):
        for i, entry in enumerate(structure):
            structure[i] = validate_structure(entry, level=level + 1)
        return structure
    if isinstance(structure, dict):
        for k in structure:
            structure[k] = validate_structure(structure[k], level=level + 1)
        return structure
    raise ValueError(f"Unsupported structure {structure}")


class DependencyCollection:
    def __init__(self, structure, keyword=None):
        self.structure = validate_structure(structure)
        self.keyword = keyword
        self._index = None
        self._re_index()

    def __getitem__(self, key: int):
        if self._index is None:
            raise ValueError("Index not built.")
        return self.get(self._index[key])

    def __setitem__(self, key: int, value):
        if self._index is None:
            raise ValueError("Index not built.")
        return self.set(self._index[key], value)

    def __len__(self):
        if self._index is None:
            raise ValueError("Index not built.")
        return len(self._index)

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def index(self, value):
        for i in range(len(self)):
            if self[i] == value:
                return i
        return -1

    def get(self, multi_index):
        e = self.structure
        for j in multi_index:
            e = e[j]  # type: ignore
        return e

    def set(self, multi_index, value):
        e = self.structure
        for i, j in enumerate(multi_index):
            if i == len(multi_index) - 1:
                e[j] = value  # type: ignore

    def append(self, value, flex_key=None, index=None):
        if self._index is None:
            raise ValueError("Index not built.")
        i = len(self._index)
        if isinstance(self.structure, list):
            if index is not None:
                self.structure.insert(index, value)
                self._re_index()
                return index
            else:
                self.structure.append(value)
                self._re_index()
                return i
        if isinstance(self.structure, dict):
            flex_key = f"{DEPENDENCY_APPEND_PREFIX}{i}" if flex_key is None else flex_key
            self.structure[flex_key] = value
            self._re_index()
            return flex_key

    def _re_index(self):
        self._index = build_index(self.structure, [], [])


# endregion

# region Dash blueprint


def collect_args(args: Union[Tuple[Any], List[Any]], inputs, outputs):
    for arg in args:
        if isinstance(arg, (list, tuple)):
            collect_args(arg, inputs, outputs)
            continue
        if isinstance(arg, Output):
            outputs.append(arg)
            continue
        if isinstance(arg, (Input, State)):
            inputs.append(arg)
            continue
        # If we get here, the argument was not recognized.
        raise ValueError(f"Unsupported argument: {arg}")
    return DependencyCollection(inputs), DependencyCollection(outputs)


class DummyDependency(DashDependency):
    """
    Object used to represent dummy dependencies.
    """

    def __init__(self, component_id, component_property):
        super().__init__(component_id, component_property)


class CallbackBlueprint:
    def __init__(self, *args, **kwargs):
        # Collect args "normally".
        self.inputs, self.outputs = collect_args(args, [], [])
        # Flexible signature handling via keyword arguments. If provided, it takes precedence.
        if "output" in kwargs:
            self.outputs = DependencyCollection(kwargs.pop("output"), keyword="output")
        if "inputs" in kwargs:
            self.inputs = DependencyCollection(kwargs.pop("inputs"), keyword="inputs")
        if "state" in kwargs:
            raise ValueError("Please use the 'inputs' keyword instead of the 'state' keyword.")
        # Collect dummy elements.
        if kwargs.get("background", False) and "progress" in kwargs:
            # This element represents the set_progress function.
            self.inputs.append(
                DummyDependency("function", "set_progress"),
                index=0,
                flex_key="set_progress",
            )
        # Collect the rest.
        self.kwargs: Dict[str, Any] = kwargs
        self.f = None

    def register(self, app: dash.Dash):  # noqa: C901
        # Collect dependencies.
        dep_args, dep_kwargs = [], {}
        for dep_col in [self.outputs, self.inputs]:
            s = dep_col.structure
            if isinstance(s, list) and len(s) == 1:
                s = s[0]
            if isinstance(s, DummyDependency):
                continue
            if isinstance(s, list):
                s = [dep for dep in s if not isinstance(dep, DummyDependency)]
            if dep_col.keyword is None:
                dep_args.append(s)
            else:
                dep_kwargs[dep_col.keyword] = s
        # Do binding.
        if isinstance(self.f, (str, ClientsideFunction)):
            app.clientside_callback(self.f, *dep_args, **dep_kwargs, **self.kwargs)
        else:
            app.callback(*dep_args, **dep_kwargs, **self.kwargs)(self.f)

    @property
    def uid(self) -> str:
        if isinstance(self.f, (ClientsideFunction, str)):
            f_repr = repr(self.f)  # handles clientside functions
        else:
            f_repr = f"{self.f.__module__}.{self.f.__name__}"  # handles Python functions
        f_hash = hashlib.md5(f_repr.encode()).digest()
        return str(uuid.UUID(bytes=f_hash, version=4))

    @property
    def multi_output(self) -> bool:
        # Check if there are more than  1 output.
        if len(self.outputs) > 1:
            return True
        # Check for wild card output(s).
        component_id = self.outputs[0].component_id
        if not isinstance(component_id, dict):
            return False
        # The ALL and ALLSMALLER flags indicate multi output.
        return any([component_id[k] in [ALLSMALLER, ALL] for k in component_id])


class DashBlueprint:
    def __init__(
        self,
        transforms: List[DashTransform] = None,
        include_global_callbacks: bool = False,
    ):
        self.callbacks: List[CallbackBlueprint] = []
        self.clientside_callbacks: List[CallbackBlueprint] = []
        self.transforms = _resolve_transforms(transforms)
        self._layout = None
        self._layout_is_function = False
        self.include_global_callbacks = include_global_callbacks

    def callback(self, *args, **kwargs):
        """
        This method saves the callback on the DashBlueprint object.
        """
        cbp = CallbackBlueprint(*args, **kwargs)
        self.callbacks.append(cbp)

        def wrapper(f):
            cbp.f = f
            return f

        return wrapper

    def clientside_callback(self, clientside_function, *args, **kwargs):
        """
        This method saves the clientside callback on the DashBlueprint object.
        """
        cbp = CallbackBlueprint(*args, **kwargs)
        cbp.f = clientside_function
        self.clientside_callbacks.append(cbp)

    def register_callbacks(self, app: Union[dash.Dash, DashBlueprint]):
        """
        This function registers all callbacks collected by the blueprint onto a Dash (or DashBlueprint) object.
        """
        callbacks, clientside_callbacks = self._resolve_callbacks()
        # Move callbacks from one blueprint to another.
        if isinstance(app, DashProxy):
            app.blueprint.callbacks += callbacks
            app.blueprint.clientside_callbacks += clientside_callbacks
            return
        # Register callbacks on the "real" app object.
        for cbp in callbacks + clientside_callbacks:
            cbp.register(app)

    def _resolve_callbacks(
        self,
    ) -> Tuple[List[CallbackBlueprint], List[CallbackBlueprint]]:
        """
        This method resolves the callbacks, i.e. it applies the callback injections.
        """
        callbacks, clientside_callbacks = self.callbacks, self.clientside_callbacks
        # Add any global callbacks.
        if self.include_global_callbacks:
            callbacks += GLOBAL_BLUEPRINT.callbacks
            clientside_callbacks += GLOBAL_BLUEPRINT.clientside_callbacks
        # Proceed as before.
        for transform in self.transforms:
            callbacks, clientside_callbacks = transform.apply(callbacks, clientside_callbacks)
        return callbacks, clientside_callbacks

    # TODO: Include or not? The plugin still seems a bit immature.
    def register(
        self,
        app: Union[dash.Dash, DashProxy],
        module,
        prefix: Union[str, PrefixIdTransform, None] = None,
        **kwargs,
    ):
        # Add prefix transform if supplied.
        if prefix is not None:
            prefix_transform = prefix if isinstance(prefix, PrefixIdTransform) else PrefixIdTransform(prefix)
            self.transforms.append(prefix_transform)
        # Register the callbacks and page.
        self.register_callbacks(app)
        dash.register_page(module, layout=self._layout_value, **kwargs)

    def clear(self):
        self.callbacks = []
        self.clientside_callbacks = []
        self.transforms = []

    def _layout_value(self, *args, **kwargs):
        layout = self._layout(*args, **kwargs) if self._layout_is_function else self._layout
        for transform in self.transforms:
            layout = transform.layout(layout, self._layout_is_function)
        return layout

    def embed(self, app: Union[DashBlueprint, DashProxy]):
        if isinstance(app, DashBlueprint):
            self.register_callbacks(app)
            return self._layout_value()
        if app.blueprint._layout_is_function and app._got_first_request["setup_server"]:
            return self._layout_value()
        self.register_callbacks(app)
        return self._layout_value()

    @property
    def layout(self):
        return self._layout

    @layout.setter
    def layout(self, value):
        self._layout_is_function = isinstance(value, patch_collections_abc("Callable"))
        self._layout = value


# endregion

# region Dash proxy


class DashProxy(dash.Dash):
    """
    DashProxy is a wrapper around the DashBlueprint object enabling drop-in replacement of the original Dash object. It
    enables transforms (via the DashBlueprint object), performs the necessary app configuration for all transforms to
    work (e.g. setting a secret key on the server), and exposes convenience functions such as 'hijack'.
    """

    def __init__(
        self,
        *args,
        transforms=None,
        include_global_callbacks=True,
        blueprint=None,
        prevent_initial_callbacks="initial_duplicate",
        **kwargs,
    ):
        super().__init__(*args, prevent_initial_callbacks=prevent_initial_callbacks, **kwargs)
        self.blueprint = (
            DashBlueprint(transforms, include_global_callbacks=include_global_callbacks)
            if blueprint is None
            else blueprint
        )
        self.setup_server_lock = threading.Lock()

    def callback(self, *args, **kwargs):
        return self.blueprint.callback(*args, **kwargs)

    def clientside_callback(self, clientside_function, *args, **kwargs):
        return self.blueprint.clientside_callback(clientside_function, *args, **kwargs)

    def long_callback(self, *_args, **_kwargs):
        raise NotImplementedError(
            "The 'long_callback(..)' syntax is not supported, please use 'callback(background=True, ...)' instead."
        )

    def register_celery_tasks(self):
        if sys.argv[0].endswith("celery"):
            self.register_callbacks()

    def register_callbacks(self):
        self.blueprint.register_callbacks(super())

    def _setup_server(self):
        with self.setup_server_lock:
            first_request = not bool(self._got_first_request["setup_server"])
            if first_request:
                # Trigger callback generation for embedded layouts.
                if self.blueprint._layout_is_function:
                    _ = self.blueprint.layout()
                self.register_callbacks()
            # Proceed as normally.
            super()._setup_server()
            if first_request:
                # Remap callback bindings to enable callback registration via the 'before_first_request' hook.
                self.callback = super().callback
                self.clientside_callback = super().clientside_callback
                # Set session secret. Used by some subclasses.
                if not self.server.secret_key:
                    self.server.secret_key = secrets.token_urlsafe(16)

    def hijack(self, app: dash.Dash):
        """
        Hijack another app. Typically, used with Dataiku 10 where the Dash object is instantiated outside user code.
        """
        # Change properties.
        readonly_props = app.config.__dict__.get("_read_only", {})
        app.config.update({k: v for k, v in self.config.items() if k not in readonly_props})
        app.title = self.title
        app.index_string = self.index_string
        # Inject layout.
        app.layout = html.Div()  # fool layout validator
        app._layout_value = self._layout_value
        # Register callbacks.
        self.blueprint.register_callbacks(app)
        # Setup secret.
        if not app.server.secret_key:
            app.server.secret_key = secrets.token_urlsafe(16)

    def _layout_value(self):
        return self.blueprint._layout_value()

    @property
    def layout(self):
        return self.blueprint._layout

    @layout.setter
    def layout(self, value):
        self.blueprint.layout = value


def _get_session_id(session_key=None):
    session_key = "session_id" if session_key is None else session_key
    # Create unique session id.
    if not session.get(session_key):
        session[session_key] = secrets.token_urlsafe(16)
    return session.get(session_key)


def _extract_list_from_kwargs(kwargs: dict, key: str) -> list:
    if kwargs is not None and key in kwargs:
        contents = kwargs.pop(key)
        if contents is None:
            return []
        if isinstance(contents, list):
            return contents
        else:
            return [contents]
    else:
        return []


# endregion

# region Dash transform


class DashTransform:
    def __init__(self):
        self.layout_initialized = False

    def apply(self, callbacks, clientside_callbacks):
        return self.apply_serverside(callbacks), self.apply_clientside(clientside_callbacks)

    def apply_serverside(self, callbacks):
        return callbacks  # per default do nothing

    def apply_clientside(self, callbacks):
        return callbacks  # per default do nothing

    def layout(self, layout, layout_is_function):
        if layout_is_function or not self.layout_initialized:
            self.transform_layout(layout)
            self.layout_initialized = True
        return layout

    def transform_layout(self, layout):
        return layout  # per default, do nothing

    def get_dependent_transforms(self):
        return []

    def sort_key(self):
        return 1


class StatefulDashTransform(DashTransform):
    def __init__(self):
        super().__init__()
        self.blueprint = DashBlueprint()
        self.components = []


def _resolve_transforms(transforms: List[DashTransform]) -> List[DashTransform]:
    # Resolve transforms.
    transforms = [] if transforms is None else transforms
    dependent_transforms = []
    for transform in transforms:
        for dependent_transform in transform.get_dependent_transforms():
            # Check if the dependent transform is already there.
            if any([isinstance(t, dependent_transform.__class__) for t in transforms]):
                continue
                # Otherwise, add it.
            dependent_transforms.append(dependent_transform)
    return sorted(transforms + dependent_transforms, key=lambda t: t.sort_key())


# endregion

# region Blocking callback transform


class BlockingCallbackTransform(StatefulDashTransform):
    def __init__(self, timeout=60):
        super().__init__()
        self.timeout = timeout

    def transform_layout(self, layout):
        children = _as_list(layout.children) + self.components
        layout.children = children

    def apply(self, callbacks, clientside_callbacks):
        callbacks = self.apply_serverside(callbacks)
        return callbacks, clientside_callbacks + self.blueprint.clientside_callbacks

    def apply_serverside(self, callbacks):
        for callback in callbacks:
            if not callback.kwargs.get("blocking", None):
                continue

            timeout = callback.kwargs.get("blocking_timeout", self.timeout)
            callback_id = callback.uid
            # Bind proxy components.
            start_client_id = f"{callback_id}_start_client"
            start_client_ctx = f"{callback_id}_start_client_ctx"
            end_server_id = f"{callback_id}_end_server"
            end_client_id = f"{callback_id}_end_client"
            start_blocked_id = f"{callback_id}_start_blocked"
            end_blocked_id = f"{callback_id}_end_blocked"
            self.components.extend(
                [
                    dcc.Store(id=start_client_id),
                    dcc.Store(id=end_server_id),
                    dcc.Store(id=end_client_id),
                    dcc.Store(id=start_blocked_id),
                    dcc.Store(id=start_client_ctx),
                    CycleBreaker(id=end_blocked_id),
                ]
            )
            # Bind start signal callback.
            start_callback = f"""function()
            {{
                const start = arguments[arguments.length-3];
                const end = arguments[arguments.length-2];
                let ctx = arguments[arguments.length-1];
                const now = new Date().getTime();
                const trigger = dash_clientside.callback_context.triggered[0];
                const no = window.dash_clientside.no_update
                // Update context.
                if(trigger !== undefined){{
                    if(!trigger.prop_id.startsWith('{end_blocked_id}')){{
                        ctx = {{}}
                        keys = ["inputs", "inputs_list", "triggered"];
                        for (let i = 0; i < keys.length; i++) {{
                            let key = keys[i];
                            ctx[key] = dash_clientside.callback_context[key];
                        }}
                    }}
                }}
                // First run => INVOKE.
                if(!end & !start){{
                    return [now, null, ctx];
                }}
                // Timeout reached  => INVOKE (but don't refresh context).
                if((now - start)/1000 > {timeout}){{
                    return [now, null, ctx];
                }}
                // Not completed first time => BLOCK.
                if(!end){{
                    return [no, now, no];
                }}
                // Previous invoke ended => INVOKE (but don't update context).
                if(end > start){{
                    return [now, null, ctx];
                }}
                // Callback running.
                return [no, now, no];
            }}"""
            self.blueprint.clientside_callback(
                start_callback,
                [
                    Output(start_client_id, "data"),
                    Output(start_blocked_id, "data"),
                    Output(start_client_ctx, "data"),
                ],
                [Input(end_blocked_id, "dst")] + list(callback.inputs),
                [
                    State(start_client_id, "data"),
                    State(end_client_id, "data"),
                    State(start_client_ctx, "data"),
                ],
            )
            # Bind end signal callback.
            end_callback = """function(endServerId, startBlockedId)
            {
                const now = new Date().getTime();
                if(startBlockedId){
                    return [now, now];
                }
                return [now, window.dash_clientside.no_update];
            }"""
            self.blueprint.clientside_callback(
                end_callback,
                [Output(end_client_id, "data"), Output(end_blocked_id, "src")],
                Input(end_server_id, "data"),
                State(start_blocked_id, "data"),
            )
            # Modify the original callback to send finished signal.
            num_outputs = len(callback.outputs)
            out_flex_key = callback.outputs.append(Output(end_server_id, "data"))
            # Change original inputs to state.
            for i, item in enumerate(callback.inputs):
                callback.inputs[i] = State(item.component_id, item.component_property)
            # Add new input trigger.
            in_flex_key = callback.inputs.append(Input(start_client_id, "data"))
            st_flex_key = callback.inputs.append(State(start_client_ctx, "data"))
            # Modify the callback function accordingly.
            f = callback.f
            callback.f = skip_input_signal_add_output_signal(num_outputs, out_flex_key, in_flex_key, st_flex_key)(f)

        return callbacks


def skip_input_signal_add_output_signal(num_outputs, out_flex_key, in_flex_key, st_flex_key):
    def wrapper(f):
        @functools.wraps(f)
        def decorated_function(*args, **kwargs):
            args, kwargs, fltr = _skip_inputs(args, kwargs, [in_flex_key, st_flex_key])
            cached_ctx = fltr[1]
            single_output = num_outputs <= 1
            if cached_ctx is not None and "triggered" in cached_ctx:
                local_ctx = context_value.get()
                local_ctx["triggered_inputs"] = cached_ctx["triggered"]
                context_value.set(local_ctx)
            try:
                outputs = f(*args, **kwargs)
            except Exception:
                logging.exception(f"Exception raised in blocking callback [{f.__name__}]")
                outputs = no_update if single_output else [no_update] * num_outputs

            return _append_output(outputs, datetime.utcnow().timestamp(), single_output, out_flex_key)

        return decorated_function

    return wrapper


# endregion

# region Log transform


class LogConfig:
    def __init__(
        self,
        log_output,
        log_writer_map: Dict[int, Callable],
        layout_transform: Callable[[List[Component]], List[Component]],
    ):
        self.log_output = log_output
        self.log_writer_map = log_writer_map
        self.layout_transform = layout_transform


def setup_notifications_log_config():
    log_id = "notifications_provider"
    log_output = Output(log_id, "children", allow_duplicate=True)

    def notification_layout_transform(layout: List[Component]):
        import dash_mantine_components as dmc

        layout.append(html.Div(id=log_id))
        if dmc.__version__ < "0.14.0":
            layout.append(dmc.NotificationsProvider())
        else:
            layout.append(dmc.NotificationProvider(zIndex=2000))

        return layout

    return LogConfig(log_output, get_notification_log_writers(), notification_layout_transform)


def setup_div_log_config():
    log_id = "log"

    def div_layout_transform(layout: List[Component]):
        layout.append(html.Div(id=log_id))
        return layout

    log_output = Output(log_id, "children", allow_duplicate=True)
    return LogConfig(log_output, get_default_log_writers(), div_layout_transform)


def get_default_log_writers():
    return {
        logging.INFO: lambda x, **kwargs: html.Div(f"INFO: {x}", **kwargs),
        logging.WARNING: lambda x, **kwargs: html.Div(f"WARNING: {x}", **kwargs),
        logging.ERROR: lambda x, **kwargs: html.Div(f"ERROR: {x}", **kwargs),
    }


def get_notification_log_writers():
    import dash_mantine_components as dmc

    def _default_kwargs(color, title, message):
        return dict(
            color=color,
            title=title,
            message=message,
            id=str(uuid.uuid4()),
            action="show",
            autoClose=False,
        )

    def log_info(message, **kwargs):
        return dmc.Notification(**{**_default_kwargs("blue", "Info", message), **kwargs})

    def log_warning(message, **kwargs):
        return dmc.Notification(**{**_default_kwargs("yellow", "Warning", message), **kwargs})

    def log_error(message, **kwargs):
        return dmc.Notification(**{**_default_kwargs("red", "Error", message), **kwargs})

    return {
        logging.INFO: log_info,
        logging.WARNING: log_warning,
        logging.ERROR: log_error,
    }


class DashLogger:
    def __init__(self, log_writers: Dict[int, Callable]):
        self.log_writers = log_writers
        self.output = []

    def clear(self):
        self.output.clear()

    def info(self, message, **kwargs):
        self.log(logging.INFO, message, **kwargs)

    def warning(self, message, **kwargs):
        self.log(logging.WARNING, message, **kwargs)

    def error(self, message, **kwargs):
        self.log(logging.ERROR, message, **kwargs)

    def log(self, level, message, **kwargs):
        self.output.append(self.log_writers[level](message, **kwargs))

    def get_output(self):
        return self.output if self.output else dash.no_update


class LogTransform(DashTransform):
    def __init__(self, log_config=None, try_use_mantine=True):
        super().__init__()
        # Per default, try to use dmc notification system.
        if log_config is None and try_use_mantine:
            try:
                log_config = setup_notifications_log_config()
            except ImportError:
                msg = "Failed to import dash-mantine-components, falling back to simple div for log output."
                logging.warning(msg)
        # Otherwise, use simple div.
        if log_config is None:
            log_config = setup_div_log_config()
        # Bind the resulting log config.
        self.log_config = log_config

    def transform_layout(self, layout):
        layout.children = self.log_config.layout_transform(_as_list(layout.children))

    def apply(self, callbacks, clientside_callbacks):
        callbacks = self.apply_serverside(callbacks)
        return callbacks, clientside_callbacks

    def apply_serverside(self, callbacks):
        for callback in callbacks:
            if not callback.kwargs.get("log", None):
                continue
            # Add the log component as output.
            single_output = len(callback.outputs) <= 1
            out_flex_key = callback.outputs.append(self.log_config.log_output)
            # Modify the callback function accordingly.
            f = callback.f
            logger = DashLogger(self.log_config.log_writer_map)  # TODO: What about scope?
            callback.f = bind_logger(logger, single_output, out_flex_key)(f)

        return callbacks

    def get_dependent_transforms(self):
        return [MultiplexerTransform()]


def bind_logger(logger, single_output, out_flex_key):
    def wrapper(f):
        @functools.wraps(f)
        def decorated_function(*args, **kwargs):
            logger.clear()
            outputs = f(*args, **kwargs, dash_logger=logger)
            return _append_output(outputs, logger.get_output(), single_output, out_flex_key)

        return decorated_function

    return wrapper


# endregion

# region Loading transform


class LoadingTransform(DashTransform):
    """
    The loading transform makes it easy to trigger a single loading component from multiple callbacks.
    """

    def __init__(self, **kwargs):
        super().__init__()
        kwargs_defaults = {
            "overlay_style": {
                "opacity": 0.5,
                "backgroundColor": "white",
                "width": "100vw",
                "height": "100vh",
                "position": "fixed",
                "left": 0,
                "top": 0,
                "zIndex": 1000,
                "visibility": "visible",
            },
            "delay_show": 500,
            "delay_hide": 1000,
            "fullscreen": True,
            "id": "loading_id",
        }
        self.kwargs = {**kwargs_defaults, **kwargs}

    def transform_layout(self, layout):
        loading = dcc.Loading(html.Div(""), **self.kwargs)
        layout.children = _as_list(layout.children) + [loading]

    def apply(self, callbacks, clientside_callbacks):
        callbacks = self.apply_serverside(callbacks)
        return callbacks, clientside_callbacks

    def apply_serverside(self, callbacks):
        for callback in callbacks:
            if not callback.kwargs.get("loading", None):
                continue
            # Add the log component as output.
            single_output = len(callback.outputs) <= 1
            out_flex_key = callback.outputs.append(Output(self.kwargs["id"], "children", allow_duplicate=True))
            # Modify the callback function accordingly.
            f = callback.f
            callback.f = bind_loading(single_output, out_flex_key)(f)

        return callbacks


def bind_loading(single_output, out_flex_key):
    def wrapper(f):
        @functools.wraps(f)
        def decorated_function(*args, **kwargs):
            outputs = f(*args, **kwargs)
            return _append_output(outputs, dash.no_update, single_output, out_flex_key)

        return decorated_function

    return wrapper


# endregion

# region Cycle breaker transform


class CycleBreakerTransform(StatefulDashTransform):
    def __init__(self):
        super().__init__()

    def transform_layout(self, layout):
        children = _as_list(layout.children) + self.components
        layout.children = children

    def apply(self, callbacks, clientside_callbacks):
        cycle_inputs = {}
        # Update inputs.
        for c in callbacks + clientside_callbacks:
            for i in c.inputs:
                if isinstance(i, CycleBreakerInput):
                    cid = self._cycle_break_id(i)
                    cycle_inputs[cid] = (i.component_id, i.component_property)
                    i.component_id = cid
                    i.component_property = "dst"
        # Construct components.
        self.components = [CycleBreaker(id=cid) for cid in cycle_inputs]
        # Construct callbacks.
        f = "function(x){return x;}"
        cycle_callbacks = []
        for cid in cycle_inputs:
            cb = CallbackBlueprint(Output(cid, "src"), Input(*cycle_inputs[cid]))
            cb.f = f
            cycle_callbacks.append(cb)
        return callbacks, clientside_callbacks + cycle_callbacks

    @staticmethod
    def _cycle_break_id(d: DashDependency):
        return f"{str(d).replace('.', '_')}_breaker"


class CycleBreakerInput(Input):
    def __init__(self, component_id, component_property):
        super().__init__(component_id, component_property)


# endregion

# region Global blueprint object (to emulate Dash 2.0 import syntax)

GLOBAL_BLUEPRINT = DashBlueprint()


def callback(*args, **kwargs):
    return GLOBAL_BLUEPRINT.callback(*args, **kwargs)


def clientside_callback(clientside_function, *args, **kwargs):
    return GLOBAL_BLUEPRINT.clientside_callback(clientside_function, *args, **kwargs)


# TODO: Include or not? The plugin still seems a bit immature.
def register(blueprint: DashBlueprint, name: str, prefix=None, **kwargs):
    if prefix is not None:
        prefix_transform = PrefixIdTransform(prefix)
        blueprint.transforms.append(prefix_transform)
    blueprint.register_callbacks(GLOBAL_BLUEPRINT)
    dash.register_page(name, layout=blueprint._layout_value, **kwargs)


# endregion

# region Prefix ID transform


class PrefixIdTransform(DashTransform):
    def __init__(self, prefix, prefix_func=None, escape=None):
        """
        The PrefixIdTransform adds a prefix to all component ids of the DashBlueprint, including
        their references in callbacks. It is typically used to avoid ID collisions between
        blueprints when they are registered on or embedded in the main Dash application.

        Args:
            prefix (str): The prefix that will be added to the component_ids.
            prefix_func (callable(str, Component, callable(str)): A function used to modify
                the ID of the components. This function must accept three arguments: the
                prefix string, a Dash component, and the escape function. If not provided,
                the `prefix_component()` function is used, which relies on `apply_prefix()`
                to calculate the new ID.
            escape (callable(str)): A function that will be called by `apply_prefix()` to
                determine whether the component's ID should remain unaltered (escaped). The
                function should accept a string (the component_id) and return a boolean.
                By default, `default_prefix_escape()` is used, which avoids modifying
                certain IDs that start with "a-" or "anchor-".

        Note: `PrefixIdTransform` is automatically registered as `transforms` by
               the `DashBlueprint.register()` method when the `prefix` parameter is specified.
        """
        super().__init__()
        self.prefix = prefix
        self.prefix_func = prefix_func if prefix_func is not None else prefix_component
        self.escape = default_prefix_escape if escape is None else escape

    def _apply(self, callbacks):
        for callback in callbacks:
            for i in callback.inputs:
                i.component_id = apply_prefix(self.prefix, i.component_id, self.escape)
            for o in callback.outputs:
                o.component_id = apply_prefix(self.prefix, o.component_id, self.escape)
        return callbacks

    def apply_serverside(self, callbacks):
        return self._apply(callbacks)

    def apply_clientside(self, callbacks):
        return self._apply(callbacks)

    def transform_layout(self, layout):
        prefix_recursively(layout, self.prefix, self.prefix_func, self.escape)


def default_prefix_escape(component_id: str):
    if isinstance(component_id, str):
        if component_id.startswith("a-"):  # intended usage is for anchors
            return True
        if component_id.startswith("anchor-"):  # intended usage is for anchors
            return True
    return False


def apply_prefix(prefix, component_id, escape):
    if escape(component_id):
        return component_id
    if isinstance(component_id, dict):
        for key in component_id:
            # This branch handles the IDs. TODO: Can we always assume use of ints?
            if isinstance(component_id[key], int):
                continue
            # This branch handles the wildcard callbacks.
            if isinstance(component_id[key], _Wildcard):
                continue
            # All "normal" props are prefixed.
            component_id[key] = "{}-{}".format(prefix, component_id[key])
        return component_id
    return "{}-{}".format(prefix, component_id)


def prefix_recursively(item, key, prefix_func, escape):
    prefix_func(key, item, escape)
    if hasattr(item, "children"):
        children = _as_list(item.children)
        for child in children:
            prefix_recursively(child, key, prefix_func, escape)


def prefix_component(key: str, component: Component, escape: Callable):
    if hasattr(component, "id"):
        component.id = apply_prefix(key, component.id, escape)
    if not hasattr(component, "_namespace"):
        return
    # Special handling of dash bootstrap components. TODO: Maybe add others?
    if component._namespace == "dash_bootstrap_components":
        if component._type == "Tooltip":
            component.target = apply_prefix(key, component.target, escape)


# TODO: Test this one.
def dynamic_prefix(app: Union[DashBlueprint, DashProxy], component: Component):
    bp: DashBlueprint = app if isinstance(app, DashBlueprint) else app.blueprint
    prefix_transforms = list(filter(lambda t: isinstance(t, PrefixIdTransform), bp.transforms))
    # No transform, just return.
    if len(prefix_transforms) == 0:
        return
    prefix_transform: PrefixIdTransform = prefix_transforms[0]
    prefix_recursively(
        component,
        prefix_transform.prefix,
        prefix_transform.prefix_func,
        prefix_transform.escape,
    )


# endregion

# region Trigger transform (the only default transform)


class Trigger(Input):
    """
    Like an Input, a trigger can trigger a callback, but it's values it not included in the resulting function call.
    """

    def __init__(self, component_id, component_property):
        super().__init__(component_id, component_property)


class TriggerTransform(DashTransform):
    def apply_serverside(self, callbacks):
        for callback in callbacks:
            is_trigger = [isinstance(item, Trigger) for item in callback.inputs]
            # Check if any triggers are there.
            if not any(is_trigger):
                continue
            # If so, filter the callback args.
            f = callback.f
            callback.f = filter_args(is_trigger)(f)
        return callbacks

    def apply_clientside(self, callbacks):
        for callback in callbacks:
            is_not_trigger = [not isinstance(item, Trigger) for item in callback.inputs]
            # Check if any triggers are there.
            if all(is_not_trigger):
                continue
            # If so, filter the callback args.
            args = [f"arg{i}" for i in range(len(callback.inputs))]
            filtered_args = compress(args, is_not_trigger)
            if isinstance(callback.f, ClientsideFunction):
                callback.f = f"window.dash_clientside['{callback.f.namespace}']['{callback.f.function_name}']"
            callback.f = f"""
function({", ".join(args)}) {{
const func = {callback.f};
return func({", ".join(filtered_args)});
}}"""
        return callbacks


def filter_args(args_filter):
    def wrapper(f):
        @functools.wraps(f)
        def decorated_function(*args):
            post_args = list(args[len(args_filter) :])
            args = list(args[: len(args_filter)])
            filtered_args = [arg for j, arg in enumerate(args) if not args_filter[j]] + post_args
            return f(*filtered_args)

        return decorated_function

    return wrapper


def trigger_filter(args):
    inputs_args = [item for item in args if isinstance(item, Input) or isinstance(item, State)]
    is_trigger = [isinstance(item, Trigger) for item in inputs_args]
    return is_trigger


# endregion

# region Multiplexer transform


class MultiplexerTransform(DashTransform):
    """
    The MultiplexerTransform was previously used to make it possible to target an output by multiple callbacks, but as
    per Dash 2.9 this function is now included, but by default disabled. To achieve similar behaviour as before (and
    thus improve backwards compatibility), the MultiplexerTransform enables the functionality automagically.
    """

    def __init__(self):
        super().__init__()

    def apply(self, callbacks, clientside_callbacks):
        all_callbacks = callbacks + clientside_callbacks
        # Group by output.
        output_map = defaultdict(list)
        for callback in all_callbacks:
            for output in callback.outputs:
                output_map[_output_id_without_wildcards(output)].append(output)
        # Set allow_duplicate where needed.
        for output_id in output_map:
            # If there is only one output, multiplexing is not needed.
            if len(output_map[output_id]) == 1:
                continue
            for entry in output_map[output_id]:
                entry.allow_duplicate = True

        return callbacks, clientside_callbacks


def _output_id_without_wildcards(output: Output) -> str:
    i, p = output.component_id, output.component_property
    if isinstance(i, dict):
        i = json.dumps({k: i[k] for k in sorted(i) if i[k] not in [ALL, MATCH, ALLSMALLER]})
    return f"{i}_{p}"


# endregion

# region SerializationTransform


class SerializationTransform(DashTransform):
    def apply_serverside(self, callbacks):
        for callback in callbacks:
            f = callback.f
            callback.f = self._unpack_pack_callback(callback)(f)
        return callbacks

    def _try_load(self, data: Any, ann=None):
        raise NotImplementedError

    def _try_dump(self, obj: Any):
        raise NotImplementedError

    def _unpack_pack_callback(self, callback):  # noqa: C901
        full_arg_spec = inspect.getfullargspec(callback.f)

        def unpack_pack_args(f):
            @functools.wraps(f)
            def decorated_function(*args, **kwargs):
                args = list(args)
                # Replace args and kwargs. # TODO: Is recursion needed?
                for i, arg in enumerate(args):
                    an = full_arg_spec.annotations.get(full_arg_spec.args[i])
                    value = [self._try_load(a, an) for a in arg] if isinstance(arg, list) else self._try_load(arg, an)
                    args[i] = value
                for key in kwargs:
                    arg = kwargs[key]
                    an = full_arg_spec.annotations.get(key)
                    value = [self._try_load(a, an) for a in arg] if isinstance(arg, list) else self._try_load(arg, an)
                    kwargs[key] = value
                # Evaluate function.
                data = f(*args, **kwargs)
                # Capture outputs.
                data = self._try_dump(data)
                if isinstance(data, list):
                    data = [self._try_dump(element) for element in data]
                if isinstance(data, tuple):
                    data = tuple([self._try_dump(element) for element in data])
                if isinstance(data, dict):
                    data = {key: self._try_dump(data[key]) for key in data}
                return data

            return decorated_function

        return unpack_pack_args

    def sort_key(self):
        return 0


# endregion

# region DataclassTransform


class DataclassTransform(SerializationTransform):
    def _try_load(self, data: Any, ann=None) -> Any:
        if not dataclasses.is_dataclass(ann):
            return data
        if data is None:
            return None
        if isinstance(data, str):
            data = json.loads(data)
        if isinstance(data, dict):
            return fromdict(ann, data)
        raise ValueError(f"Unsupported data type for dataclass: {type(data)}")

    def _try_dump(self, obj: Any) -> Any:
        if not dataclasses.is_dataclass(obj):
            return obj
        return asdict(obj)


# endregion


# region PydanticTransform


class BaseModelTransform(SerializationTransform):
    def _try_load(self, data: Any, ann=None) -> Any:
        if not isinstance(ann, type(BaseModel)):
            return data
        if data is None:
            return None
        if isinstance(data, str):
            return ann.model_validate_json(data)
        if isinstance(data, dict):
            return ann.model_validate(data)
        raise ValueError(f"Unsupported data type for Pydantic model: {type(data)}")

    def _try_dump(self, obj: Any) -> Any:
        if not isinstance(obj, BaseModel):
            return obj
        return obj.model_dump_json()


# endregion

# region Server side output transform


class ServersideBackend:
    def get(self, key, ignore_expired=False):
        raise NotImplementedError()

    def set(self, key, value):
        raise NotImplementedError()

    def has(self, key):
        raise NotImplementedError()

    @property
    def uid(self) -> str:
        """
        Backend identifier. Must be unique across the backend registry. Defaults to class name.
        """
        return self.__class__.__name__


class FileSystemBackend(FileSystemCache, ServersideBackend):
    def __init__(self, cache_dir="file_system_backend", **kwargs):
        super().__init__(cache_dir, **kwargs)

    def get(self, key: str, ignore_expired=False):
        if key is None:
            return None
        if not ignore_expired:
            return super().get(key)
        # TODO: This part must be implemented for each type of cache.
        filename = self._get_filename(key)
        try:
            with self._safe_stream_open(filename, "rb") as f:
                _ = struct.unpack("I", f.read(4))[0]
                return self.serializer.load(f)
        except FileNotFoundError:
            pass
        except (OSError, EOFError, struct.error):
            logging.warning(
                "Exception raised while handling cache file '%s'",
                filename,
                exc_info=True,
            )
        return None

    @property
    def uid(self) -> str:
        """
        Backend identifier. Must be unique across the backend registry.
        """
        return f"{self.__class__.__name__}:{self._path}"


class RedisBackend(RedisCache, ServersideBackend):
    """
    Store that uses Redis as backend. Note, that the timeout must be large enough that a (k,v) pair NEVER expires
    during a user session. If it does, the user experience for those sessions will be degraded.
    """

    def __init__(self, default_timeout=24 * 3600, **kwargs):
        super().__init__(default_timeout=default_timeout, **kwargs)

    def get(self, key, ignore_expired=False):
        # TODO: Is there any way to honor ignore_expired for redis? I don't think so
        return super().get(key)


class EnrichedOutput(Output):
    """
    Like a normal Output, includes additional properties related to storing the data.
    """

    def __init__(
        self,
        component_id,
        component_property,
        allow_duplicate=False,
        backend=None,
        session_check=None,
        arg_check=True,
    ):
        super().__init__(component_id, component_property, allow_duplicate)
        self.backend = backend
        self.session_check = session_check
        self.arg_check = arg_check


class ServersideOutputTransform(SerializationTransform):
    prefix: str = "SERVERSIDE_"

    def __init__(
        self,
        backends: Optional[List[ServersideBackend]] = None,
        default_backend: Optional[ServersideBackend] = None,
    ):
        super().__init__()
        # Per default, use file system backend.
        if backends is None:
            backends = [FileSystemBackend()]
        self._default_backend: ServersideBackend = backends[0] if default_backend is None else default_backend
        # Setup registry for easy/fast access.
        self._backend_registry: Dict[str, ServersideBackend] = {backend.uid: backend for backend in backends}

    def _try_load(self, data: Any, ann=None) -> Any:
        if not isinstance(data, str):
            return data
        if not data.startswith(self.prefix):
            return data
        obj = json.loads(data[len(self.prefix) :])
        backend = self._backend_registry[obj["backend_uid"]]
        value = backend.get(obj["key"], ignore_expired=True)
        return value

    def _try_dump(self, obj: Any) -> Any:
        if not isinstance(obj, Serverside):
            return obj
        backend_uid = obj.backend_uid
        # If not backend it set, use the default.
        if backend_uid is None:
            backend_uid = self._default_backend.uid
        # Dump the data.
        backend = self._backend_registry[backend_uid]
        backend.set(obj.key, obj.value)
        # Return lookup structure.
        data = dict(backend_uid=backend_uid, key=obj.key)
        return f"{self.prefix}{json.dumps(data)}"


class Serverside(Generic[T]):
    def __init__(
        self,
        value: T,
        key: str = None,
        backend: Union[ServersideBackend, str, None] = None,
    ):
        self.value = value
        self.key: str = str(uuid.uuid4()) if key is None else key
        self.backend_uid: str = backend.uid if isinstance(backend, ServersideBackend) else backend


# endregion

# region No output transform


class NoOutputTransform(StatefulDashTransform):
    def __init__(self):
        super().__init__()

    def transform_layout(self, layout):
        children = _as_list(layout.children) + self.components
        layout.children = children

    def _apply(self, callbacks):
        for callback in callbacks:
            if len(callback.outputs) == 0:
                output_id = callback.uid
                hidden_div = html.Div(id=output_id, style={"display": "none"})
                callback.outputs.append(Output(output_id, "children"))
                self.components.append(hidden_div)
        return callbacks

    def apply_serverside(self, callbacks):
        return self._apply(callbacks)

    def apply_clientside(self, callbacks):
        return self._apply(callbacks)

    def sort_key(self):
        return 0


# endregion

# region Batteries included dash proxy object


class Dash(DashProxy):
    def __init__(self, *args, **kwargs):
        transforms = [
            TriggerTransform(),
            LogTransform(),
            MultiplexerTransform(),
            NoOutputTransform(),
            CycleBreakerTransform(),
            BlockingCallbackTransform(),
            ServersideOutputTransform(),
        ]
        super().__init__(*args, transforms=transforms, **kwargs)


# endregion

# region Utils


def _as_list(item):
    if item is None:
        return []
    if isinstance(item, tuple):
        return list(item)
    if isinstance(item, list):
        return item
    return [item]


def _skip_inputs(args, kwargs, keys: List[Any]):
    fltr = []
    str_keys = [key for key in keys if isinstance(key, str)]
    int_keys = [key for key in keys if isinstance(key, int)]
    # Drop str keys.
    for key in str_keys:
        fltr.append(kwargs.pop(key))
    # Filter the other ones.
    if any(int_keys):
        args = list(args)
        fltr.extend([arg for i, arg in enumerate(args) if i in keys])
        args = [arg for i, arg in enumerate(args) if i not in keys]
    # TODO: Do we need to handle None?
    return args, kwargs, fltr


def _append_output(outputs, value, single_output, out_idx):
    # Handle flex signature.
    if isinstance(outputs, dict):
        outputs[out_idx] = value
        return outputs
    # Handle single output.
    if single_output:
        return [outputs, value]
    # Finally, the "normal" case.
    return _as_list(outputs) + [value]


def _create_callback_id(item):
    cid = item.component_id
    if isinstance(cid, dict):
        cid = {key: cid[key] for key in cid if cid[key] not in _wildcard_mappings}
        cid = json.dumps(cid)
    return "{}.{}".format(cid, item.component_property)


def _check_multi(item):
    cid = item.component_id
    if not isinstance(cid, dict):
        return False
    vs = cid.values()
    return ALL in vs or ALLSMALLER in vs


def plotly_jsonify(data):
    return json.loads(json.dumps(data, cls=plotly.utils.PlotlyJSONEncoder))


# endregion



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class Loading(Component):
    """A Loading component.
The Loading component makes it possible to stop event propagation during loading.

Keyword arguments:

- children (list of a list of or a singular dash component, string or numbers | a list of or a singular dash component, string or number; optional):
    Array that holds components to render.

- id (string; optional):
    The ID of this component, used to identify dash components in
    callbacks. The ID needs to be unique across all of the components
    in an app.

- loading_state (dict; optional):
    Object that holds the loading state object coming from
    dash-renderer.

    `loading_state` is a dict with keys:

    - component_name (string; optional):
        Holds the name of the component that is loading.

    - is_loading (boolean; optional):
        Determines if the component is loading or not.

    - prop_name (string; optional):
        Holds which property is loading.

- preventDefault (list of strings; default ["keydown"]):
    Events for which to call preventDefault() during loading."""
    _children_props = []
    _base_nodes = ['children']
    _namespace = 'dash_extensions'
    _type = 'Loading'
    @_explicitize_args
    def __init__(self, children=None, id=Component.UNDEFINED, preventDefault=Component.UNDEFINED, loading_state=Component.UNDEFINED, **kwargs):
        self._prop_names = ['children', 'id', 'loading_state', 'preventDefault']
        self._valid_wildcard_attributes =            []
        self.available_properties = ['children', 'id', 'loading_state', 'preventDefault']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs and excess named props
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}

        super(Loading, self).__init__(children=children, **args)



import json
from typing import List, Union

from dash import Input, Output, callback_context, html
from dash.development.base_component import Component

# region Get triggered


class Triggered(object):
    def __init__(self, id, **kwargs):
        self.id = id
        for key in kwargs:
            setattr(self, key, kwargs[key])


def get_triggered() -> Triggered:  # noqa: C901
    triggered = callback_context.triggered
    if not triggered:
        return Triggered(None)
    # Collect trigger ids and values.
    triggered_id = None
    triggered_values = {}
    for entry in triggered:
        # TODO: Test this part.
        elements = entry["prop_id"].split(".")
        current_id = ".".join(elements[:-1])
        current_prop = elements[-1]
        # Determine the trigger object.
        if triggered_id is None:
            triggered_id = current_id
        # TODO: Should all properties of the trigger be registered, or only one?
        if triggered_id != current_id:
            continue
        triggered_values[current_prop] = entry["value"]
    # Now, create an object.
    try:
        triggered_id = json.loads(triggered_id)
    except ValueError:
        pass
    return Triggered(triggered_id, **triggered_values)


# endregion

# region Utils for html tables

Node = Union[str, float, int, Component]


def generate_html_table(
    columns: List[Node],
    rows: List[List[Node]] = None,
    footers: List[Node] = None,
    caption: Node = None,
) -> List[Component]:
    rows = [] if rows is None else rows
    # Create table structure.
    html_header = [html.Tr([html.Th(col) for col in columns])]
    html_rows = [html.Tr([html.Td(children=cell) for cell in row]) for row in rows]
    html_table = [html.Thead(html_header), html.Tbody(html_rows)]
    # Add (optional) caption.
    if caption is not None:
        html_table = [html.Caption(caption)] + html_table
    # Add (optional) footer.
    if footers is not None:
        html_footer = [html.Tr([html.Th(col) for col in footers])]
        html_table += html.Tfoot(html_footer)
    return html_table


# endregion


def fix_page_load_anchor_issue(app, delay, input_id=None, output_id=None):
    """
    Fixes the issue that the pages is not scrolled to the anchor position on initial load.
    :param app: the Dash app object
    :param delay: in some cases, an additional delay might be needed for the page to load, specify in ms
    :param input_id: id of input dummy element
    :param output_id: id of output dummy element
    :return: dummy elements, which must be added to the layout for the fix to work
    """
    # Create dummy components.
    input_id = input_id if input_id is not None else "fix_page_load_anchor_issue_input"
    output_id = output_id if output_id is not None else "fix_page_load_anchor_issue_output"
    dummy_input = html.Div(id=input_id, style={"display": "hidden"})
    dummy_output = html.Div(id=output_id, style={"display": "hidden"})
    # Setup the callback that does the magic.
    app.clientside_callback(
        """
        function(dummy_value) {{
            setTimeout(function(){{
                const match = document.getElementById(window.location.hash.substring(1))
                match.scrollIntoView();
            }}, {});
        }}
        """.format(delay),
        Output(output_id, "children"),
        [Input(input_id, "children")],
        prevent_initial_call=False,
    )
    return [dummy_input, dummy_output]



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class Keyboard(Component):
    """A Keyboard component.
The Keyboard component listens for keyboard events.

Keyword arguments:

- children (a list of or a singular dash component, string or number; optional):
    The children of this component. If any children are provided, the
    component will listen for events from these     components. If no
    children are specified, the component will listen for events from
    the document object.

- id (string; optional):
    The ID used to identify this component in Dash callbacks.

- captureKeys (list of string | dicts; optional):
    The keys to capture. Defaults to all keys. Can be either a string
    (e.g. \"Enter\") or an object (e.g. \"{key: 'Enter', ctrlKey:
    True}\").

- className (string; optional):
    A custom class name.

- eventProps (list of strings; default ["key", "altKey", "ctrlKey", "shiftKey", "metaKey", "repeat"]):
    The event properties to forward to dash, see
    https://developer.mozilla.org/en-US/docs/Web/API/KeyboardEvent.

- keydown (dict; optional):
    keydown (dict) the object that holds the result of the key down
    event. It is a dictionary with the following keys:      \"key\",
    \"altKey\", \"ctrlKey\", \"shiftKey\",\"metaKey\", \"repeat\".
    Those keys have the following values:     - key (str) which key is
    pressed    - altKey (bool) whether the Alt key is pressed    -
    ctrlKey (bool) Ctrl key is pressed    - shiftKey (bool) Shift key
    is pressed    - metaKey (bool) Meta key is pressed (Mac: Command
    key or PC: Windows key)    - repeat (bool) whether the key is held
    down.

- keys_pressed (dict; optional):
    keys_pressed (dict) is a dict of objects like keydown for all keys
    currently pressed.

- keyup (dict; optional):
    keyup (dict) the object that holds the result of the key up event.
    Structure like keydown.

- n_keydowns (number; default 0):
    A counter, which is incremented on each key down event, similar to
    n_clicks for buttons.

- n_keyups (number; default 0):
    A counter, which is incremented on each key up event, similar to
    n_clicks for buttons.

- style (dict; optional):
    The CSS style of the component.

- useCapture (boolean; default False):
    Value of useCapture used when registering event listeners."""
    _children_props = []
    _base_nodes = ['children']
    _namespace = 'dash_extensions'
    _type = 'Keyboard'
    @_explicitize_args
    def __init__(self, children=None, style=Component.UNDEFINED, className=Component.UNDEFINED, id=Component.UNDEFINED, eventProps=Component.UNDEFINED, captureKeys=Component.UNDEFINED, keydown=Component.UNDEFINED, keyup=Component.UNDEFINED, keys_pressed=Component.UNDEFINED, n_keydowns=Component.UNDEFINED, n_keyups=Component.UNDEFINED, useCapture=Component.UNDEFINED, **kwargs):
        self._prop_names = ['children', 'id', 'captureKeys', 'className', 'eventProps', 'keydown', 'keys_pressed', 'keyup', 'n_keydowns', 'n_keyups', 'style', 'useCapture']
        self._valid_wildcard_attributes =            []
        self.available_properties = ['children', 'id', 'captureKeys', 'className', 'eventProps', 'keydown', 'keys_pressed', 'keyup', 'n_keydowns', 'n_keyups', 'style', 'useCapture']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs and excess named props
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}

        super(Keyboard, self).__init__(children=children, **args)



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class CycleBreaker(Component):
    """A CycleBreaker component.
Simple data store that automatically copies the current value of the src property into dst property. Can be used to break circular dependencies.

Keyword arguments:

- id (string; optional):
    The ID used to identify this component in Dash callbacks.

- dst (boolean | number | string | dict | list; optional):
    Read the forwarded value from this property.

- src (boolean | number | string | dict | list; optional):
    Set this property to value to be forwarded from ."""
    _children_props = []
    _base_nodes = ['children']
    _namespace = 'dash_extensions'
    _type = 'CycleBreaker'
    @_explicitize_args
    def __init__(self, id=Component.UNDEFINED, src=Component.UNDEFINED, dst=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'dst', 'src']
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'dst', 'src']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs and excess named props
        args = {k: _locals[k] for k in _explicit_args}

        super(CycleBreaker, self).__init__(**args)



import dash.dependencies

# region Random component id generation

_set_random_id = dash.development.base_component.Component._set_random_id
_components_with_random_ids = []


def _collect_components_with_random_ids(self):
    if not hasattr(self, "id"):
        _components_with_random_ids.append(self)
    return _set_random_id(self)


dash.development.base_component.Component._set_random_id = _collect_components_with_random_ids


def assert_no_random_ids():
    """
    When a component is referenced in a callback, a random id is assigned automatically, if an id is not already set.
    This operation makes the application _stateful_, which is *not* recommended. This function raises an assertion
    error if any components have been assigned random ids, thereby forcing the developer to set ids for all components.
    """
    if _components_with_random_ids:
        return
    raise AssertionError(
        f"The following components have random ids: {', '.join([str(c) for c in _components_with_random_ids])}"
    )


# endregion



import os
import re
from distutils.dir_util import copy_tree

from dash import Dash

"""
The purpose of this module is to ease the integration of Dash with Dataiku.
"""


def bind_assets_folder(app: Dash, app_id: str, assets_folder: str):  # noqa: C901
    """
    Dataiku 10 doesn't support separate asset folders for each Dash app. This function targets fixing this issue by
    (1) creating a new asset sub folder for each app, and (2) limiting asset loading to this folder.
    """
    # Create assets container folder for the app.
    dst_assets_folder = os.path.join(app.config.assets_folder, app_id)
    os.makedirs(dst_assets_folder, exist_ok=True)
    # Copy assets.
    copy_tree(assets_folder, dst_assets_folder)

    def _walk_assets_directory(self):
        walk_dir = os.path.join(self.config.assets_folder, app_id)  # EMHER: Use application sub dir
        slash_splitter = re.compile(r"[\\/]+")
        ignore_str = self.config.assets_ignore
        ignore_filter = re.compile(ignore_str) if ignore_str else None

        for current, _, files in sorted(os.walk(walk_dir)):
            if current == walk_dir:
                base = ""
            else:
                s = current.replace(walk_dir, "").lstrip("\\").lstrip("/")
                splitted = slash_splitter.split(s)
                if len(splitted) > 1:
                    base = "/".join(slash_splitter.split(s))
                else:
                    base = splitted[0]
            base = os.path.join(base, app_id)  # EMHER: Use application sub dir
            if ignore_filter:
                files_gen = (x for x in files if not ignore_filter.search(x))
            else:
                files_gen = files

            for f in sorted(files_gen):
                path = "/".join([base, f]) if base else f

                full = os.path.join(current, f)

                if f.endswith("js"):
                    self.scripts.append_script(self._add_assets_resource(path, full))
                elif f.endswith("css"):
                    self.css.append_css(self._add_assets_resource(path, full))
                elif f == "favicon.ico":
                    self._favicon = path

    app._walk_assets_directory = lambda x=app: _walk_assets_directory(app)



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class SSE(Component):
    """A SSE component.
The SSE component makes it possible to collect data from e.g. a ResponseStream. It's a wrapper around the SSE.js library.
https://github.com/mpetazzoni/sse.js

Keyword arguments:

- id (string; optional):
    The ID used to identify this component in Dash callbacks.

- animate_chunk (number; default 1):
    Chunk size (i.e. number of characters) for the animation.

- animate_delay (number; default 0):
    If set, each character is delayed by some amount of time. Used to
    animate the stream.

- animate_prefix (string; optional):
    Prefix to be excluded from the animation.

- animate_suffix (string; optional):
    Suffix to be excluded from the animation.

- animation (string; optional):
    The data value. Either the latest, or the concatenated depending
    on the `concat` property.

- concat (boolean; default True):
    A boolean indicating if the stream values should be concatenated.

- done (boolean; optional):
    A boolean indicating if the (current) stream has ended.

- options (dict; optional):
    Options passed to the SSE constructor.
    https://github.com/mpetazzoni/sse.js?tab=readme-ov-file#options-reference.

    `options` is a dict with keys:

    - debug (boolean; optional)

    - headers (dict; optional)

    - method (string; optional)

    - payload (dict; optional)

    - start (boolean; optional)

    - withCredentials (boolean; optional)

- url (string; optional):
    URL of the endpoint.

- value (string; optional):
    The data value. Either the latest, or the concatenated depending
    on the `concat` property."""
    _children_props = []
    _base_nodes = ['children']
    _namespace = 'dash_extensions'
    _type = 'SSE'
    @_explicitize_args
    def __init__(self, id=Component.UNDEFINED, options=Component.UNDEFINED, url=Component.UNDEFINED, concat=Component.UNDEFINED, value=Component.UNDEFINED, animate_delay=Component.UNDEFINED, animate_chunk=Component.UNDEFINED, animate_prefix=Component.UNDEFINED, animate_suffix=Component.UNDEFINED, animation=Component.UNDEFINED, done=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'animate_chunk', 'animate_delay', 'animate_prefix', 'animate_suffix', 'animation', 'concat', 'done', 'options', 'url', 'value']
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'animate_chunk', 'animate_delay', 'animate_prefix', 'animate_suffix', 'animation', 'concat', 'done', 'options', 'url', 'value']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs and excess named props
        args = {k: _locals[k] for k in _explicit_args}

        super(SSE, self).__init__(**args)



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class BeforeAfter(Component):
    """A BeforeAfter component.
Before After Image Slider based on https://github.com/sneas/img-comparison-slider

Keyword arguments:

- id (string; optional):
    The ID used to identify this component in Dash callbacks.

- after (dict; optional):
    Props for the `after` Img component. eg {\"src\":
    \"/assets/lena_color.png\"}.

- before (dict; optional):
    Props for the `before` Img component. eg {\"src\":
    \"/assets/lena_bw.png\"}.

- direction (a value equal to: 'horizontal', 'vertical'; default 'horizontal'):
    Set slider direction.

- height (string; default 'auto'):
    Image height - default \"auto\" for responsive images.

- hover (boolean; default True):
    Automatic slide on mouse over.

- keyboard (a value equal to: 'enabled', 'disabled'; default 'enabled'):
    Enable/disable slider position control with the keyboard.

- value (number; default 50):
    The divider position can be specified as a percentage, i.e. 0 to
    100.

- width (string; default '100%'):
    Image width - default \"100%\" for responsive images."""
    _children_props = []
    _base_nodes = ['children']
    _namespace = 'dash_extensions'
    _type = 'BeforeAfter'
    @_explicitize_args
    def __init__(self, id=Component.UNDEFINED, height=Component.UNDEFINED, width=Component.UNDEFINED, hover=Component.UNDEFINED, value=Component.UNDEFINED, direction=Component.UNDEFINED, keyboard=Component.UNDEFINED, before=Component.UNDEFINED, after=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'after', 'before', 'direction', 'height', 'hover', 'keyboard', 'value', 'width']
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'after', 'before', 'direction', 'height', 'hover', 'keyboard', 'value', 'width']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs and excess named props
        args = {k: _locals[k] for k in _explicit_args}

        super(BeforeAfter, self).__init__(**args)



import os

import jsbeautifier

# region Templates

_template = """window.{namespace} = Object.assign({{}}, window.{namespace}, {{
    {content}
}});"""
_ns_template = """{namespace}: {{
        {content}
    }}"""
_func_template = """{name}: {function}"""


# endregion


class Namespace:
    def __init__(self, *args):
        self.args = list(args)
        self.f_map = {}

    def __call__(self, var):
        all_args = self.args + [var]
        return variable(*all_args)

    def add(self, src, name=None):
        # Check if the function already exists.
        for key in self.f_map:
            if name is None and self.f_map[key] == src:
                return key
        # Otherwise add it.
        name = f"function{len(self.f_map)}" if name is None else name
        self.f_map[name] = src
        return name

    def dump(self, assets_folder="assets"):
        os.makedirs(assets_folder, exist_ok=True)
        content = "\n,".join([_func_template.format(name=name, function=self.f_map[name]) for name in self.f_map])
        for ns in reversed(self.args[1:]):
            content = _ns_template.format(namespace=ns, content=content)
        content = _template.format(namespace=self.args[0], content=content)
        with open(os.path.join(assets_folder, "{}.js".format("_".join(self.args))), "w") as f:
            f.write(jsbeautifier.beautify(content))


def assign(src, name=None):
    name = _default_name_space.add(src, name)
    _default_name_space.dump()
    return _default_name_space(name)


def arrow_function(value):
    return dict(arrow=value)


def variable(*args):
    return dict(variable=".".join(list(args)))


_default_name_space = Namespace("dashExtensions", "default")



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class Purify(Component):
    """A Purify component.
The Html component makes it possible to render html sanitized via DOMPurify.

Keyword arguments:

- id (string; optional):
    The ID used to identify this component in Dash callbacks.

- className (string; optional):
    The class of the component.

- config (dict; optional):
    Configuration (optional) of DOMPurify, see the docs
    https://github.com/cure53/DOMPurify.

- html (string; optional):
    Html string."""
    _children_props = []
    _base_nodes = ['children']
    _namespace = 'dash_extensions'
    _type = 'Purify'
    @_explicitize_args
    def __init__(self, html=Component.UNDEFINED, config=Component.UNDEFINED, id=Component.UNDEFINED, className=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'className', 'config', 'html']
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'className', 'config', 'html']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs and excess named props
        args = {k: _locals[k] for k in _explicit_args}

        super(Purify, self).__init__(**args)



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class EventListener(Component):
    """An EventListener component.
The EventListener component listens for events from the document object or children if provided.

Keyword arguments:

- children (a list of or a singular dash component, string or number; optional):
    The children of this component. If any children are provided, the
    component will listen for events from these      components. If no
    children are specified, the component will listen for events from
    the document object.

- id (string; optional):
    The ID used to identify this component in Dash callbacks.

- className (string; optional):
    A custom class name.

- event (dict; optional):
    The latest event fired.

- events (list of dicts; default [{"event": "keydown", "props": ["key", "altKey", "ctrlKey", "shiftKey","metaKey", "repeat"]}]):
    The event entry specifies which event to listen to, e.g. \"click\"
    for click events. The \"props\" entry specifies      what event
    properties to record, e.g. [\"x\", \"y\"] to get the cursor
    position.

    `events` is a list of dicts with keys:

    - event (string; optional)

    - props (list of strings; optional)

- logging (boolean; default False):
    If True, event information is logged to the javascript console.
    Useful if you can't remember events props.

- n_events (number; default 0):
    The number of events fired.

- style (dict; optional):
    The CSS style of the component.

- useCapture (boolean; default False):
    Value of useCapture used when registering event listeners."""
    _children_props = []
    _base_nodes = ['children']
    _namespace = 'dash_extensions'
    _type = 'EventListener'
    @_explicitize_args
    def __init__(self, children=None, id=Component.UNDEFINED, events=Component.UNDEFINED, logging=Component.UNDEFINED, style=Component.UNDEFINED, className=Component.UNDEFINED, event=Component.UNDEFINED, n_events=Component.UNDEFINED, useCapture=Component.UNDEFINED, **kwargs):
        self._prop_names = ['children', 'id', 'className', 'event', 'events', 'logging', 'n_events', 'style', 'useCapture']
        self._valid_wildcard_attributes =            []
        self.available_properties = ['children', 'id', 'className', 'event', 'events', 'logging', 'n_events', 'style', 'useCapture']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs and excess named props
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}

        super(EventListener, self).__init__(children=children, **args)



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class Lottie(Component):
    """A Lottie component.


Keyword arguments:

- children (a list of or a singular dash component, string or number; optional):
    The children of this component.

- id (string; optional):
    The ID used to identify this component in Dash callbacks.

- ariaLabel (string; optional)

- ariaRole (string; optional)

- className (string; optional):
    The class of the component.

- direction (number; optional)

- height (string; optional):
    Pixel value for containers height.

- isClickToPauseDisabled (boolean; optional)

- isPaused (boolean; optional)

- isStopped (boolean; optional)

- options (dict; optional):
    Options passed to the Lottie animation (see
    https://www.npmjs.com/package/react-lottie for details).

- segments (list of numbers; optional)

- speed (number; optional)

- style (string; optional)

- title (string; optional)

- url (string; optional):
    If set, data will be downloaded from this url.

- width (string; optional):
    Pixel value for containers width."""
    _children_props = []
    _base_nodes = ['children']
    _namespace = 'dash_extensions'
    _type = 'Lottie'
    @_explicitize_args
    def __init__(self, children=None, id=Component.UNDEFINED, className=Component.UNDEFINED, options=Component.UNDEFINED, url=Component.UNDEFINED, width=Component.UNDEFINED, height=Component.UNDEFINED, isStopped=Component.UNDEFINED, isPaused=Component.UNDEFINED, speed=Component.UNDEFINED, segments=Component.UNDEFINED, direction=Component.UNDEFINED, ariaRole=Component.UNDEFINED, ariaLabel=Component.UNDEFINED, isClickToPauseDisabled=Component.UNDEFINED, title=Component.UNDEFINED, style=Component.UNDEFINED, **kwargs):
        self._prop_names = ['children', 'id', 'ariaLabel', 'ariaRole', 'className', 'direction', 'height', 'isClickToPauseDisabled', 'isPaused', 'isStopped', 'options', 'segments', 'speed', 'style', 'title', 'url', 'width']
        self._valid_wildcard_attributes =            []
        self.available_properties = ['children', 'id', 'ariaLabel', 'ariaRole', 'className', 'direction', 'height', 'isClickToPauseDisabled', 'isPaused', 'isStopped', 'options', 'segments', 'speed', 'style', 'title', 'url', 'width']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs and excess named props
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}

        super(Lottie, self).__init__(children=children, **args)



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class Rating(Component):
    """A Rating component.
Rating component based on https://github.com/voronianski/react-star-rating-component

Keyword arguments:

- id (string; optional):
    The ID used to identify this component in Dash callbacks.

- activeColor (string; optional):
    This is the color of the icon in the active state.

- count (number; optional):
    The number of Icons to display.

- defaultColor (string; optional):
    This is the color of the icon in the inactive state.

- defaultValue (string; optional):
    Required. This is the value of the rating displayed by default.
    Supply this if your rating is also a readOnly.

- readOnly (boolean; default True):
    This sets the component to be non editable.

- shape (string; optional):
    This is the shape displayed as icon.

- size (string; optional):
    This defines the size of the Icons used.

- spacing (string; optional):
    This defines the fap between the Icons used."""
    _children_props = []
    _base_nodes = ['children']
    _namespace = 'dash_extensions'
    _type = 'Rating'
    @_explicitize_args
    def __init__(self, id=Component.UNDEFINED, defaultValue=Component.UNDEFINED, count=Component.UNDEFINED, shape=Component.UNDEFINED, readOnly=Component.UNDEFINED, size=Component.UNDEFINED, spacing=Component.UNDEFINED, activeColor=Component.UNDEFINED, defaultColor=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'activeColor', 'count', 'defaultColor', 'defaultValue', 'readOnly', 'shape', 'size', 'spacing']
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'activeColor', 'count', 'defaultColor', 'defaultValue', 'readOnly', 'shape', 'size', 'spacing']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs and excess named props
        args = {k: _locals[k] for k in _explicit_args}

        super(Rating, self).__init__(**args)



from .BeforeAfter import BeforeAfter
from .CycleBreaker import CycleBreaker
from .DeferScript import DeferScript
from .EventListener import EventListener
from .EventSource import EventSource
from .Keyboard import Keyboard
from .Loading import Loading
from .Lottie import Lottie
from .Mermaid import Mermaid
from .Purify import Purify
from .Rating import Rating
from .SSE import SSE
from .WebSocket import WebSocket

__all__ = [
    "BeforeAfter",
    "CycleBreaker",
    "DeferScript",
    "EventListener",
    "EventSource",
    "Keyboard",
    "Loading",
    "Lottie",
    "Mermaid",
    "Purify",
    "Rating",
    "SSE",
    "WebSocket"
]


# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class WebSocket(Component):
    """A WebSocket component.
A simple interface to

Keyword arguments:

- id (string; optional):
    The ID used to identify this component in Dash callbacks.

- error (dict | string; optional):
    This property is set with the content of the onerror event.

- message (dict | string; optional):
    When messages are received, this property is updated with the
    message content.

- protocols (list of strings; optional):
    Supported websocket protocols (optional).

- send (dict | string; optional):
    When this property is set, a message is sent with its content.

- state (dict | string; default {readyState: WebSocket.CONNECTING}):
    This websocket state (in the readyState prop) and associated
    information.

- url (string; optional):
    The websocket endpoint (e.g. wss://echo.websocket.org)."""
    _children_props = []
    _base_nodes = ['children']
    _namespace = 'dash_extensions'
    _type = 'WebSocket'
    @_explicitize_args
    def __init__(self, state=Component.UNDEFINED, message=Component.UNDEFINED, error=Component.UNDEFINED, send=Component.UNDEFINED, url=Component.UNDEFINED, protocols=Component.UNDEFINED, id=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'error', 'message', 'protocols', 'send', 'state', 'url']
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'error', 'message', 'protocols', 'send', 'state', 'url']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs and excess named props
        args = {k: _locals[k] for k in _explicit_args}

        super(WebSocket, self).__init__(**args)
